@@ -1,8 +1,8 @@
 ;;; func-menu.el         --- Jump to a function within a buffer.
 ;;;
 ;;; David Hughes <d.hughes@videonetworks.com>
-;;; Last Modified: David Hughes 5th March 1998
-;;; Version: 2.54
+;;; Last Modified: David Hughes 13th July 1998
+;;; Version: 2.57
 ;;; Keywords: tools, c, lisp
 ;;;
 ;;; This program is free software; you can redistribute it and/or modify
@@ -51,13 +51,25 @@
 ;;;
 ;;; Modes supported:
 ;;; ================
-;;; Ada, Assembly, Bacis2, BibTex, C++, C, Dired, Ehdm, ELisp, FORTRAN, Ksh,
-;;; Latex, Lelisp, Makefile, Maple, Modula2, Modula3, Outline, Objective-C,
-;;; Pascal, Perl, Postscript, Prolog, PVS, Python, SGML, Scheme, Tcl, Verilog
+;;; Ada, Assembly, BibTex, C++, C, Dired, Ehdm, ELisp, FORTRAN, Ksh, Latex,
+;;; Lelisp, Makefile, Maple, Modula2, Modula3, Outline, Objective-C, Pascal,
+;;; Perl, Postscript, Prolog, PVS, Python, SGML, Scheme, Tcl, Verilog
 ;;;
 ;;; Acknowledgements:
 ;;; =================
 ;;;
+;;; Protect integer concatenation with int-to-string
+;;; Anthony Rossini <arossini@biostat.washington.edu>
+;;;
+;;; Patch for jde-mode
+;;; Tony Lam <tonyl@Eng.Sun.COM>
+;;;
+;;; Patches to FORTRAN support
+;;; Bruce Ravel <ravel@phys.washington.edu>
+;;;
+;;; Idea for fume-goto-function-hook
+;;; Mark Gates <mgates@ncsa.uiuc.edu>
+;;;
 ;;; Patch for backward compatibility with backquote
 ;;; Bill Dubuque <wgd@martigny.ai.mit.edu>
 ;;;
@@ -144,6 +156,7 @@
 ;;;
 ;;; Ada support
 ;;; Scott Evans  <gse@ocsystems.com>
+;;; Mike Konerman <konerman@SSA.crane.navy.mil>
 ;;; Michael Polo <mikep@polo.mn.org> <mikep@cfsmo.honeywell.com>
 ;;;
 ;;; Scheme, BibTeX, Ehdm & PVS support
@@ -209,16 +222,25 @@
 
 ;;; Code
 
+(defmacro fume-bomb-proof (&rest forms)
+  (` (condition-case () (progn (,@ forms)) (t nil))))
+
 (eval-when-compile
   (byte-compiler-options
     (optimize t)
-    (warnings (- free-vars unresolved))))
+    (warnings (- free-vars unresolved)))
+  ;; For older XEmacs, Win-Emacs
+  (cond ((not (fume-bomb-proof (require 'custom)))
+         (defmacro defgroup (sym def description &rest args)
+           (` (defvar (, sym) (, def) (, description))))
+         (defmacro defcustom (sym def description &rest args)
+           (` (defvar (, sym) (, def) (, description)))))))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;;;;;;;;;;;;;;;;;;;;;;  Environment Initialisation  ;;;;;;;;;;;;;;;;;;;;;;
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
-(defconst fume-version "2.54")
+(defconst fume-version "2.57")
 
 (defconst fume-developer "David Hughes <d.hughes@videonetworks.com>")
 
@@ -229,13 +251,6 @@
    'no-log
    (format "Func-Menu version %s,  1998 %s" fume-version fume-developer)))
 
-(defgroup fume nil
-  "Jump to a function within a buffer."
-  :tag "Func Menu"
-  :group 'tools
-  :group 'c
-  :group 'lisp)
-
 (defconst fume-running-xemacs (string-match "XEmacs\\|Lucid" emacs-version))
 
 (defmacro fume-defvar-local (var value &optional doc)
@@ -251,13 +266,13 @@ won't destroy func-menu control variables."
        (make-variable-buffer-local '(, var))
        (put '(, var) 'permanent-local t))))
 
-(defmacro fume-bomb-proof (&rest forms)
-  (` (condition-case () (progn (,@ forms)) (t nil))))
-
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;;;;;;  Backward compatibility hacks for older versions of XEmacs  ;;;;;;;
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
+(or (fboundp 'display-message)
+    (defun display-message (msgtype msg) (message msg)))
+
 (or (fboundp 'defalias)
     ;; poor man's defalias
     (defun defalias (sym newdef)
@@ -280,7 +295,7 @@ Associates the function with the current load file, if any."
   (defalias 'fume-event-window 'event-window))
 
 (or (fboundp 'shrink-window-if-larger-than-buffer)
-    ;; Win-Emacs doesn't have this goodie
+    ;; Win-Emacs doesn't have this
     (defun shrink-window-if-larger-than-buffer (&optional window reqd-height)
       "Shrink WINDOW to the smallest no of lines needed to display its buffer,
 or to optional REQUIRED-HEIGHT if and only if that is larger. Does nothing if
@@ -403,13 +418,20 @@ the buffer contains more lines than the present window height."
   (apply fume-add-submenu (fume-munge-menu-args menu-name submenu before)))
 
 (defconst fume-not-tty
-  (or (featurep 'menubar)     ;; XEmacs
-      (featurep 'menu-bar)))  ;; GNU Emacs
+  (or (and (fboundp 'device-type) (not (eq 'tty (device-type))))
+      (and (symbol-value 'window-system) t))) ; obsolete test
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;;;;;;;;;;;;;;;;;;;;;;;;  Customizable Variables  ;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
+(defgroup fume nil
+  "Jump to a function within a buffer."
+  :tag "Func Menu"
+  :group 'tools
+  :group 'c
+  :group 'lisp)
+
 (defcustom fume-auto-position-popup t
   "*Set to nil if you don't want the menu to appear in the corner of the window
 in which case it will track the mouse position instead."
@@ -525,6 +547,12 @@ element should come before the second.  The arguments are cons cells;
   "*Set non-nil if 'fume-prompt-function-goto' should jump without prompting
 when a valid default exists.")
 
+;;; Idea from Mark Gates <mgates@ncsa.uiuc.edu>
+;;;
+(defvar fume-goto-function-hook nil
+  "*If non-nil, this should be a function taking POS. This function will be
+called if fume-fn-window-position is nil or not a valid number.")
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;;;;;;;;;;;;;;;;;;;;;;;;  Buffer local variables  ;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
@@ -652,16 +680,17 @@ by func-menu, then do something along the lines of the following:
 ;;; Raymond L. Toy <toy@soho.crd.ge.com>
 ;;; Richard Cognot <cognot@elfgrc.co.uk>
 ;;; Greg Sjaardema <gdsjaar@sandia.gov>
+;;; Bruce Ravel <ravel@phys.washington.edu>
 (defvar fume-function-name-regexp-fortran
   (concat
    ;; >= six spaces
    "^      \\s-*"
    ;; type specs
-   "[a-zA-Z0-9*]*\\s-*"
+   "[a-zA-Z0-9* ]*\\s-*"
    ;; continuation lines
    "\\(\n     [^ 0]\\s-*\\)*"
    ;; function or subroutine
-   "\\(entry\\|ENTRY\\|function\\|FUNCTION\\|subroutine\\|SUBROUTINE\\)\\s-*"
+   "\\(entry\\|ENTRY\\|function\\|FUNCTION\\|subroutine\\|SUBROUTINE\\|block\\s-*data\\|BLOCK\\s-*DATA\\)\\s-*"
    ;; continuation lines
    "\\(\n     [^ 0]\\s-*\\)*"
    )
@@ -682,13 +711,6 @@ by func-menu, then do something along the lines of the following:
   "^\\s-*PROCEDURE\\s-+[A-Za-z0-9_-]+"
   "Expression to get Modula function names")
 
-;;; Bacis2
-;;;
-;;; CV MEDUSA's 4th generation language
-(defvar fume-function-name-regexp-bacis
-  "module_define(!\\|define_constant(!\\|sys_sysdefine(!\\|<<dbgid +\\s-*"
-  "Expression to get Bacis2 function names")
-
 ;;; Maple
 ;;;
 ;;; Luc Tancredi <Luc.Tancredi@sophia.inria.fr>
@@ -844,9 +866,10 @@ by func-menu, then do something along the lines of the following:
 
 ;;; Ada
 ;;;
+;;; Mike Konerman <konerman@SSA.crane.navy.mil>
 ;;; Michael Polo <mikep@polo.mn.org> <mikep@cfsmo.honeywell.com>
 (defvar fume-function-name-regexp-ada
-  (cons "^[ \t]*\\(procedure\\|PROCEDURE\\|function\\|FUNCTION\\)[ \n\t]+\\([a-zA-Z0-9_]+\\|\"[^\"]\"\\)" 2)
+  (cons "^[ \t]*\\([pP][rR][oO][cC][eE][dD][uU][rR][eE]\\|[fF][uU][nN][cC][tT][iI][oO][nN]\\|[tT][aA][sS][kK][ \t\n]*[bB][oO][dD][yY]\\)[ \n\t]+\\([a-zA-Z0-9_]+\\|\"[^\"]\"\\)" 2)
   "Expression to find declaration of Ada function")
 
 ;;; ignore prototypes, 'renames', 'is new' to eliminate clutter
@@ -941,9 +964,6 @@ by func-menu, then do something along the lines of the following:
     (modula-2-mode . fume-function-name-regexp-modula)
     (modula-3-mode . fume-function-name-regexp-modula)
 
-    ;; Bacis2
-    (bacis-mode . fume-function-name-regexp-bacis)
-
     ;; Maple
     (maple-mode . fume-function-name-regexp-maple)
 
@@ -952,6 +972,7 @@ by func-menu, then do something along the lines of the following:
 
     ;; Java
     (java-mode . fume-function-name-regexp-java)
+    (jde-mode  . fume-function-name-regexp-java)
 
     ;; Python
     (alice-mode  . fume-function-name-regexp-python)
@@ -1219,10 +1240,12 @@ by func-menu, then do something along the lines of the following:
               ;; delimiters so function names that happen to contain
               ;; any of these terms are not eliminated.  The old version
               ;; would ignore "notify()" since it contained "if".
-              (or (string-match "\\`\\(if\\|switch\\|catch\\|for\\|while\\)\\'"
-                                str)
-                  ;; These constructs look like method definitions but are not
-                  (cons str beg)))
+              (or
+               (string-match
+                "\\`\\(if\\|switch\\|catch\\|for\\|while\\|synchronized\\)\\'"
+                str)
+               ;; These constructs look like method definitions but are not
+               (cons str beg)))
           (fume-find-next-java-function-name buffer)))))
 
 ;;; Specialised routine to find the next Python function
@@ -1272,7 +1295,7 @@ by func-menu, then do something along the lines of the following:
   (if (re-search-forward fume-function-name-regexp nil t)
       (let ((pos (point))
             ;; name may have "_" but must start with a letter
-            (name-regexp "\\s-+[a-zA-Z]+[_a-zA-Z0-9*]*")
+            (name-regexp "\\s-+[a-zA-Z]+[_a-zA-Z0-9* ]*")
             (eol (save-excursion (end-of-line 1) (point))))
         (skip-chars-backward " \t")
         (if (re-search-forward name-regexp eol t)
@@ -1307,23 +1330,6 @@ by func-menu, then do something along the lines of the following:
         (forward-sexp)
         (cons (buffer-substring beg (point)) beg))))
 
-;;; Specialised routine to get the next bacis2 procedure name in the buffer
-;;;
-(defun fume-find-next-bacis-function-name (buffer)
-  "Searches for the next Bacis2 function in BUFFER"
-  (set-buffer buffer)
-  (if (re-search-forward fume-function-name-regexp nil t)
-      (let ((pos (point))
-            (name (fume-bomb-proof
-                   (funcall
-                    (symbol-function (intern "focus-get-function-name"))))))
-        (if (null name)
-            (fume-find-next-bacis-function-name buffer)
-          ;; jump past possible function dbgid
-          (re-search-forward
-           (format "<<dbgid +\\s-*%s%s" name "\\s-*>>") nil t)
-          (cons name pos)))))
-
 ;;; Specialized routine to get the next Maple function name in the buffer
 ;;; Luc Tancredi <Luc.Tancredi@sophia.inria.fr>
 ;;;
@@ -1339,6 +1345,7 @@ by func-menu, then do something along the lines of the following:
 ;;; Specialised routine to get the next latex section name in the buffer
 ;;; Philippe Queinnec <queinnec@cenatls.cena.dgac.fr>
 ;;; Paolo Frasconi <paolo@mcculloch.ing.unifi.it>
+;;; Anthony Rossini <arossini@biostat.washington.edu>
 ;;;
 (defun fume-find-next-latex-section-name (buffer)
   "Searches for the next latex section in BUFFER."
@@ -1352,31 +1359,32 @@ by func-menu, then do something along the lines of the following:
                      fume-tex-section 0
                      fume-tex-subsection 0
                      fume-tex-subsubsection 0
-                     name (concat fume-tex-chapter " " (upcase name))))
+                     name (concat (int-to-string fume-tex-chapter)
+                                  " " (upcase name))))
               ((string= secname "section")
                (setq fume-tex-section (1+ fume-tex-section)
                      name (concat
                            (if (> fume-tex-chapter 0)
-                               (concat fume-tex-chapter ".") "")
-                           fume-tex-section " " name)
+                               (concat (int-to-string fume-tex-chapter) ".") "")
+                           (int-to-string fume-tex-section) " " name)
                      fume-tex-subsection 0
                      fume-tex-subsubsection 0))
               ((string= secname "subsection")
                (setq fume-tex-subsection (1+ fume-tex-subsection)
                      name (concat
                            (if (> fume-tex-chapter 0)
-                               (concat fume-tex-chapter ".") "")
-                           fume-tex-section "."
-                           fume-tex-subsection " " name)
+                               (concat (int-to-string fume-tex-chapter) ".") "")
+                           (int-to-string fume-tex-section) "."
+                           (int-to-string fume-tex-subsection) " " name)
                      fume-tex-subsubsection 0))
               ((string= secname "subsubsection")
                (setq fume-tex-subsubsection (1+ fume-tex-subsubsection)
                      name (concat
                            (if (> fume-tex-chapter 0)
-                               (concat fume-tex-chapter ".") "")
-                           fume-tex-section "."
-                           fume-tex-subsection "."
-                           fume-tex-subsubsection " " name)))
+                               (concat (int-to-string fume-tex-chapter) ".") "")
+                           (int-to-string fume-tex-section) "."
+                           (int-to-string fume-tex-subsection) "."
+                           (int-to-string fume-tex-subsubsection) " " name)))
               ((string= secname "subsubsection")
                (setq name (concat "   " name))))
         (cons name beg))))
@@ -1522,7 +1530,6 @@ by func-menu, then do something along the lines of the following:
   '((ada-mode        . fume-find-next-ada-function-name)
     (alice-mode      . fume-find-next-python-function-name)
     (asm-mode        . fume-find-next-asm-function-name)
-    (bacis-mode      . fume-find-next-bacis-function-name)
     (bibtex-mode     . fume-find-next-bibtex-citation)
     (c++-mode        . fume-match-find-next-function-name)
     (c++-c-mode      . fume-match-find-next-function-name)
@@ -1543,6 +1550,7 @@ by func-menu, then do something along the lines of the following:
     (pascal-mode     . fume-find-next-pascal-function-name)
     (perl-mode       . fume-find-next-perl-function-name)
     (java-mode       . fume-find-next-java-function-name)
+    (jde-mode        . fume-find-next-java-function-name)
     (postscript-mode . fume-find-next-postscript-function-name)
     (prolog-mode .     fume-find-next-prolog-function-name)
     (pvs-mode        . fume-find-next-pvs-entity)
@@ -1756,7 +1764,6 @@ Otherwise returns fume-function-name-regexp"
          (let ((newstr (copy-sequence str)))
            (aset newstr (1- (match-end 0)) (string-to-char "b"))
            newstr))
-        ((string-match "^/../" str) "%b") ; special for smart modeline
         (t str)))
 
 (defun fume-tickle-modeline-1 ()
@@ -2039,14 +2046,17 @@ If optional arg POPMENU is non-nil, brings up the function-menu."
           ;; possibly set mark
           (or (= orig-pos (point))
               (push-mark orig-pos (null fume-scanning-message)))
-          (if (numberp fume-fn-window-position)
-              (set-window-start
-               (selected-window)
-               (save-excursion
-                 (beginning-of-line
-                  (- 1 (min (- (window-height) 2) fume-fn-window-position)))
-                 (point)))
-            (recenter)))
+          (cond ((numberp fume-fn-window-position)
+                 (set-window-start
+                  (selected-window)
+                  (save-excursion
+                    (beginning-of-line
+                     (- 1 (min (- (window-height) 2) fume-fn-window-position)))
+                    (point))))
+                ((fboundp fume-goto-function-hook)
+                 (funcall fume-goto-function-hook pos))
+                (t
+                 (recenter))))
       (ding)
       (message "%s not found" fn)
       (function-menu))))
@@ -2207,9 +2217,10 @@ the basic menu of functions."
 ;;; Espen Skoglund <espensk@stud.cs.uit.no>
 ;;; David Hughes <d.hughes@videonetworks.com>
 ;;;
-(defun fume-prompt-function-goto (&optional other-window-p)
+(defun fume-prompt-function-goto (&optional other-window-p prompt)
   "Goto function prompted for in minibuffer (with completion).
-With prefix arg, jumps to function in a different window."
+With prefix arg, jumps to function in a different window.
+PROMPT (useful for wrapper functions) must be a format string with two %s's"
   (interactive "P")
   (let* ((default-name (fume-what-looking-at t))
          (OrigBuffer (current-buffer))
@@ -2227,7 +2238,7 @@ With prefix arg, jumps to function in a different window."
                 ""
               (let ((this-command last-command)) ; preserve last-command
                 (completing-read
-                 (format "Goto function%s%s: "
+                 (format (or prompt "Goto function%s%s: ")
                          (if other-window-p " other window" "")
                          (if default-exists-p
                              (concat " (" default-name ")")
